/* basefix.c                  tab  apr2013 */

#include "ctab.h"
#include "cmisc.h"
#include "cufio.h"

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  Fix baseline on ESI images where baselines were removed.
*/
int main(int argc, char *argv[])
{
/**/
const int headersize = 115200;
char header[115200];
char header2[115200];
char bheader[115200];
/**/
char arg[900][100];
char fitsfile[100];
char biasfile[100];
char root[100];
char newfile[100];
int pixno1,pixno2,ii,jj,narg;
int sc,ec,sr,er,nc2,nr2,nc1,nr1;
int bsc,bec,bsr,ber,bnc,bnr;
int lim1,lim2,lim3,lim4,prepix,postpix;
/**/
const int mp = 3000 * 4100;
float *img1;
float *img2;
float *bimg;
/**/

/* Allocate. */
img1= (float *)calloc(mp,sizeof(float));
img2= (float *)calloc(mp,sizeof(float));
bimg= (float *)calloc(mp,sizeof(float));

/* Copy arguments. */
cargcopy(argv, argc, &narg, arg);

/* Syntax. */
if ( narg < 1 ) {
  printf("Syntax: basefix (FITS file to fix) (full bias image FITS file)\n");
  printf("\n");
  printf("\n");
  exit(0);
}

/* Read FITS file. */
strcpy(fitsfile,arg[1]); cAddFitsExt(fitsfile);
strcpy(biasfile,arg[2]); cAddFitsExt(biasfile);
printf("Reading (file to fix): %s\n",fitsfile);
printf("Reading (refrnc bias): %s\n",biasfile);

/* Read file-to-fix FITS file. */
creadfits( fitsfile, img1, mp, header, headersize );
cGetDimensions(header,headersize,&sc,&ec,&sr,&er,&nc1,&nr1);

/* Read reference bias file. */
creadfits( biasfile, bimg, mp, bheader, headersize );
cGetDimensions(bheader,headersize,&bsc,&bec,&bsr,&ber,&bnc,&bnr);

/* Baseline or overscan region. */
prepix = cinhead("PREPIX" ,bheader, headersize);
postpix= cinhead("POSTPIX",bheader, headersize);
printf("prepix=%d  postpix=%d\n",prepix,postpix);

/* Example, if:
   PREPIX  = 12.. so beginning column range is 0 to 23 (24 pixels for both readouts)
   POSTPIX = 80.. so ending column range is 2072 to 2231 (160 pixels for both readouts)
*/
lim1 = 0;
lim2 = (prepix * 2) - 1;
lim3 = bnc - (postpix * 2);
lim4 = bnc - 1;

/* New dimensions. */
nc2 = bnc;
nr2 = bnr;
printf(" old: %5d %5d %5d %5d %5d %5d  %s\n",sc,ec,sr,er,nc1,nr1,fitsfile);
printf(" new: %5d %5d %5d %5d %5d %5d  %s\n",sc,ec,sr,er,nc2,nr2,fitsfile);
printf("bias: %5d %5d %5d %5d %5d %5d  %s\n",bsc,bec,bsr,ber,bnc,bnr,biasfile);
printf(" lim: %5d %5d %5d %5d            \n",lim1,lim2,lim3,lim4);

/* Check. */
if (nr1 != nr2) { printf("***error: number of rows do not match.\n"); exit(1); }
if (bnr != nr2) { printf("***error: bias number of rows do not match.\n"); exit(1); }

/* Setup bigger image and add in bias baseline region. */
for (ii=0; ii<nc2; ++ii) {
for (jj=0; jj<nr2; ++jj) {
  pixno2= ii + (jj*nc2);
  pixno1= (ii - (lim2+1)) + (jj * nc1);
  if ((ii >= lim1)&&(ii <= lim2)) { img2[pixno2]=bimg[pixno2]; } else {
  if ((ii >= lim3)&&(ii <= lim4)) { img2[pixno2]=bimg[pixno2]; } else {
    if ((pixno1 < 0)||(pixno1 >= (nr1*nc1))) { 
      printf("***error: pixno1 out of range.\n"); 
      printf("#@# ii=%d jj=%d  pixno1=%d\n",ii,jj,pixno1);
      exit(1); 
    }
    img2[pixno2]=img1[pixno1];
  }}
}}

/* Change new header. */
strcpy(header2,header);
cinheadset("NAXIS1",nc2,header2,headersize);
cinheadset("NAXIS2",nr2,header2,headersize);

/* Write out new bigger image. */
strcpy(root,fitsfile); ii=cindex(root,".fits"); root[ii]='\0';
sprintf(newfile,"%s-new.fits",root); printf("Write '%s'.\n",newfile);
cwritefits( newfile, img2, header2, headersize );

free(img1); free(img2); free(bimg);
return(0);
}


/*
if ( creadfitsheader(fitsfile,header,headersize) == 0) {
  fprintf(stderr,"ERROR: basefix: Cannot read file %s .\n",fitsfile); exit(0);
}
*/

